/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __param = (this && this.__param) || function (paramIndex, decorator) {
    return function (target, key) { decorator(target, key, paramIndex); }
};
import { addDisposableListener, getClientArea, getDomNodePagePosition, getTotalHeight } from '../../../../base/browser/dom.js';
import { renderIcon } from '../../../../base/browser/ui/iconLabel/iconLabels.js';
import { List } from '../../../../base/browser/ui/list/listWidget.js';
import * as arrays from '../../../../base/common/arrays.js';
import { Codicon } from '../../../../base/common/codicons.js';
import { DisposableStore } from '../../../../base/common/lifecycle.js';
import { assertType } from '../../../../base/common/types.js';
import './renameInputField.css';
import { Position } from '../../../common/core/position.js';
import { NewSymbolNameTag } from '../../../common/languages.js';
import { localize } from '../../../../nls.js';
import { IContextKeyService, RawContextKey } from '../../../../platform/contextkey/common/contextkey.js';
import { IKeybindingService } from '../../../../platform/keybinding/common/keybinding.js';
import { defaultListStyles } from '../../../../platform/theme/browser/defaultStyles.js';
import { editorWidgetBackground, inputBackground, inputBorder, inputForeground, widgetBorder, widgetShadow } from '../../../../platform/theme/common/colorRegistry.js';
import { IThemeService } from '../../../../platform/theme/common/themeService.js';
/** for debugging */
const _sticky = false;
export const CONTEXT_RENAME_INPUT_VISIBLE = new RawContextKey('renameInputVisible', false, localize('renameInputVisible', "Whether the rename input widget is visible"));
export const CONTEXT_RENAME_INPUT_FOCUSED = new RawContextKey('renameInputFocused', false, localize('renameInputFocused', "Whether the rename input widget is focused"));
let RenameInputField = class RenameInputField {
    constructor(_editor, _acceptKeybindings, _themeService, _keybindingService, contextKeyService) {
        this._editor = _editor;
        this._acceptKeybindings = _acceptKeybindings;
        this._themeService = _themeService;
        this._keybindingService = _keybindingService;
        this._disposables = new DisposableStore();
        this.allowEditorOverflow = true;
        this._visibleContextKey = CONTEXT_RENAME_INPUT_VISIBLE.bindTo(contextKeyService);
        this._focusedContextKey = CONTEXT_RENAME_INPUT_FOCUSED.bindTo(contextKeyService);
        this._editor.addContentWidget(this);
        this._disposables.add(this._editor.onDidChangeConfiguration(e => {
            if (e.hasChanged(50 /* EditorOption.fontInfo */)) {
                this._updateFont();
            }
        }));
        this._disposables.add(_themeService.onDidColorThemeChange(this._updateStyles, this));
    }
    dispose() {
        this._disposables.dispose();
        this._editor.removeContentWidget(this);
    }
    getId() {
        return '__renameInputWidget';
    }
    getDomNode() {
        if (!this._domNode) {
            this._domNode = document.createElement('div');
            this._domNode.className = 'monaco-editor rename-box';
            this._input = document.createElement('input');
            this._input.className = 'rename-input';
            this._input.type = 'text';
            this._input.setAttribute('aria-label', localize('renameAriaLabel', "Rename input. Type new name and press Enter to commit."));
            // TODO@ulugbekna: is using addDisposableListener's right way to do it?
            this._disposables.add(addDisposableListener(this._input, 'focus', () => { this._focusedContextKey.set(true); }));
            this._disposables.add(addDisposableListener(this._input, 'blur', () => { this._focusedContextKey.reset(); }));
            this._domNode.appendChild(this._input);
            this._candidatesView = new CandidatesView(this._domNode, { fontInfo: this._editor.getOption(50 /* EditorOption.fontInfo */) });
            this._label = document.createElement('div');
            this._label.className = 'rename-label';
            this._domNode.appendChild(this._label);
            this._updateFont();
            this._updateStyles(this._themeService.getColorTheme());
        }
        return this._domNode;
    }
    _updateStyles(theme) {
        var _a, _b, _c, _d;
        if (!this._input || !this._domNode) {
            return;
        }
        const widgetShadowColor = theme.getColor(widgetShadow);
        const widgetBorderColor = theme.getColor(widgetBorder);
        this._domNode.style.backgroundColor = String((_a = theme.getColor(editorWidgetBackground)) !== null && _a !== void 0 ? _a : '');
        this._domNode.style.boxShadow = widgetShadowColor ? ` 0 0 8px 2px ${widgetShadowColor}` : '';
        this._domNode.style.border = widgetBorderColor ? `1px solid ${widgetBorderColor}` : '';
        this._domNode.style.color = String((_b = theme.getColor(inputForeground)) !== null && _b !== void 0 ? _b : '');
        this._input.style.backgroundColor = String((_c = theme.getColor(inputBackground)) !== null && _c !== void 0 ? _c : '');
        // this._input.style.color = String(theme.getColor(inputForeground) ?? '');
        const border = theme.getColor(inputBorder);
        this._input.style.borderWidth = border ? '1px' : '0px';
        this._input.style.borderStyle = border ? 'solid' : 'none';
        this._input.style.borderColor = (_d = border === null || border === void 0 ? void 0 : border.toString()) !== null && _d !== void 0 ? _d : 'none';
    }
    _updateFont() {
        if (!this._input || !this._label || !this._candidatesView) {
            return;
        }
        const fontInfo = this._editor.getOption(50 /* EditorOption.fontInfo */);
        this._input.style.fontFamily = fontInfo.fontFamily;
        this._input.style.fontWeight = fontInfo.fontWeight;
        this._input.style.fontSize = `${fontInfo.fontSize}px`;
        this._candidatesView.updateFont(fontInfo);
        this._label.style.fontSize = `${this._computeLabelFontSize(fontInfo.fontSize)}px`;
    }
    _computeLabelFontSize(editorFontSize) {
        return editorFontSize * 0.8;
    }
    getPosition() {
        if (!this._visible) {
            return null;
        }
        if (!this._editor.hasModel() || // @ulugbekna: shouldn't happen
            !this._editor.getDomNode() // @ulugbekna: can happen during tests based on suggestWidget's similar predicate check
        ) {
            return null;
        }
        const bodyBox = getClientArea(this.getDomNode().ownerDocument.body);
        const editorBox = getDomNodePagePosition(this._editor.getDomNode());
        const cursorBox = this._editor.getScrolledVisiblePosition(this._position);
        this._nPxAvailableAbove = cursorBox.top + editorBox.top;
        this._nPxAvailableBelow = bodyBox.height - this._nPxAvailableAbove;
        const lineHeight = this._editor.getOption(66 /* EditorOption.lineHeight */);
        const { totalHeight: candidateViewHeight } = CandidateView.getLayoutInfo({ lineHeight });
        const positionPreference = this._nPxAvailableBelow > candidateViewHeight * 6 /* approximate # of candidates to fit in (inclusive of rename input box & rename label) */
            ? [2 /* ContentWidgetPositionPreference.BELOW */, 1 /* ContentWidgetPositionPreference.ABOVE */]
            : [1 /* ContentWidgetPositionPreference.ABOVE */, 2 /* ContentWidgetPositionPreference.BELOW */];
        return {
            position: this._position,
            preference: positionPreference,
        };
    }
    beforeRender() {
        var _a, _b;
        const [accept, preview] = this._acceptKeybindings;
        this._label.innerText = localize({ key: 'label', comment: ['placeholders are keybindings, e.g "F2 to Rename, Shift+F2 to Preview"'] }, "{0} to Rename, {1} to Preview", (_a = this._keybindingService.lookupKeybinding(accept)) === null || _a === void 0 ? void 0 : _a.getLabel(), (_b = this._keybindingService.lookupKeybinding(preview)) === null || _b === void 0 ? void 0 : _b.getLabel());
        this._domNode.style.minWidth = `250px`; // to prevent from widening when candidates come in
        this._domNode.style.maxWidth = `400px`; // TODO@ulugbekna: what if we have a very long name?
        return null;
    }
    afterRender(position) {
        if (position === null) {
            // cancel rename when input widget isn't rendered anymore
            this.cancelInput(true);
            return;
        }
        if (!this._editor.hasModel() || // shouldn't happen
            !this._editor.getDomNode() // can happen during tests based on suggestWidget's similar predicate check
        ) {
            return;
        }
        assertType(this._candidatesView);
        assertType(this._nPxAvailableAbove !== undefined);
        assertType(this._nPxAvailableBelow !== undefined);
        const inputBoxHeight = getTotalHeight(this._input);
        const labelHeight = getTotalHeight(this._label);
        let totalHeightAvailable;
        if (position === 2 /* ContentWidgetPositionPreference.BELOW */) {
            totalHeightAvailable = this._nPxAvailableBelow;
        }
        else {
            totalHeightAvailable = this._nPxAvailableAbove;
        }
        this._candidatesView.layout({ height: totalHeightAvailable - labelHeight - inputBoxHeight });
    }
    acceptInput(wantsPreview) {
        var _a;
        (_a = this._currentAcceptInput) === null || _a === void 0 ? void 0 : _a.call(this, wantsPreview);
    }
    cancelInput(focusEditor) {
        var _a;
        (_a = this._currentCancelInput) === null || _a === void 0 ? void 0 : _a.call(this, focusEditor);
    }
    focusNextRenameSuggestion() {
        var _a;
        (_a = this._candidatesView) === null || _a === void 0 ? void 0 : _a.focusNext();
    }
    focusPreviousRenameSuggestion() {
        var _a;
        if (!((_a = this._candidatesView) === null || _a === void 0 ? void 0 : _a.focusPrevious())) {
            this._input.focus();
        }
    }
    getInput(where, value, selectionStart, selectionEnd, supportPreview, candidates, token) {
        this._domNode.classList.toggle('preview', supportPreview);
        this._position = new Position(where.startLineNumber, where.startColumn);
        this._input.value = value;
        this._input.setAttribute('selectionStart', selectionStart.toString());
        this._input.setAttribute('selectionEnd', selectionEnd.toString());
        this._input.size = Math.max((where.endColumn - where.startColumn) * 1.1, 20); // determines width
        const disposeOnDone = new DisposableStore();
        candidates.then(candidates => this._showRenameCandidates(candidates, value, token));
        return new Promise(resolve => {
            this._currentCancelInput = (focusEditor) => {
                var _a;
                this._currentAcceptInput = undefined;
                this._currentCancelInput = undefined;
                (_a = this._candidatesView) === null || _a === void 0 ? void 0 : _a.clearCandidates();
                resolve(focusEditor);
                return true;
            };
            this._currentAcceptInput = (wantsPreview) => {
                assertType(this._input !== undefined);
                assertType(this._candidatesView !== undefined);
                const candidateName = this._candidatesView.focusedCandidate;
                if ((candidateName === undefined && this._input.value === value) || this._input.value.trim().length === 0) {
                    this.cancelInput(true);
                    return;
                }
                this._currentAcceptInput = undefined;
                this._currentCancelInput = undefined;
                this._candidatesView.clearCandidates();
                resolve({
                    newName: candidateName !== null && candidateName !== void 0 ? candidateName : this._input.value,
                    wantsPreview: supportPreview && wantsPreview
                });
            };
            disposeOnDone.add(token.onCancellationRequested(() => this.cancelInput(true)));
            if (!_sticky) {
                disposeOnDone.add(this._editor.onDidBlurEditorWidget(() => { var _a; return this.cancelInput(!((_a = this._domNode) === null || _a === void 0 ? void 0 : _a.ownerDocument.hasFocus())); }));
            }
            this._show();
        }).finally(() => {
            disposeOnDone.dispose();
            this._hide();
        });
    }
    _show() {
        this._editor.revealLineInCenterIfOutsideViewport(this._position.lineNumber, 0 /* ScrollType.Smooth */);
        this._visible = true;
        this._visibleContextKey.set(true);
        this._editor.layoutContentWidget(this);
        setTimeout(() => {
            this._input.focus();
            this._input.setSelectionRange(parseInt(this._input.getAttribute('selectionStart')), parseInt(this._input.getAttribute('selectionEnd')));
        }, 100);
    }
    _showRenameCandidates(candidates, currentName, token) {
        if (token.isCancellationRequested) {
            return;
        }
        // deduplicate and filter out the current value
        candidates = arrays.distinct(candidates, candidate => candidate.newSymbolName);
        candidates = candidates.filter(({ newSymbolName }) => { var _a; return newSymbolName.trim().length > 0 && newSymbolName !== ((_a = this._input) === null || _a === void 0 ? void 0 : _a.value) && newSymbolName !== currentName; });
        if (candidates.length < 1) {
            return;
        }
        // show the candidates
        this._candidatesView.setCandidates(candidates);
        // ask editor to re-layout given that the widget is now of a different size after rendering rename candidates
        this._editor.layoutContentWidget(this);
    }
    _hide() {
        this._visible = false;
        this._visibleContextKey.reset();
        this._editor.layoutContentWidget(this);
    }
};
RenameInputField = __decorate([
    __param(2, IThemeService),
    __param(3, IKeybindingService),
    __param(4, IContextKeyService)
], RenameInputField);
export { RenameInputField };
export class CandidatesView {
    constructor(parent, opts) {
        this._availableHeight = 0;
        this._lineHeight = opts.fontInfo.lineHeight;
        this._listContainer = document.createElement('div');
        this._listContainer.style.fontFamily = opts.fontInfo.fontFamily;
        this._listContainer.style.fontWeight = opts.fontInfo.fontWeight;
        this._listContainer.style.fontSize = `${opts.fontInfo.fontSize}px`;
        parent.appendChild(this._listContainer);
        const that = this;
        const virtualDelegate = new class {
            getTemplateId(element) {
                return 'candidate';
            }
            getHeight(element) {
                return that.candidateViewHeight;
            }
        };
        const renderer = new class {
            constructor() {
                this.templateId = 'candidate';
            }
            renderTemplate(container) {
                return new CandidateView(container, { lineHeight: that._lineHeight });
            }
            renderElement(candidate, index, templateData) {
                templateData.model = candidate;
            }
            disposeTemplate(templateData) {
                templateData.dispose();
            }
        };
        this._listWidget = new List('NewSymbolNameCandidates', this._listContainer, virtualDelegate, [renderer], {
            keyboardSupport: false, // @ulugbekna: because we handle keyboard events through proper commands & keybinding service, see `rename.ts`
            mouseSupport: true,
            multipleSelectionSupport: false,
        });
        this._listWidget.style(defaultListStyles);
    }
    get candidateViewHeight() {
        const { totalHeight } = CandidateView.getLayoutInfo({ lineHeight: this._lineHeight });
        return totalHeight;
    }
    // height - max height allowed by parent element
    layout({ height }) {
        this._availableHeight = height;
        if (this._listWidget.length > 0) { // candidates have been set
            this._listWidget.layout(this._pickListHeight(this._listWidget.length));
        }
    }
    _pickListHeight(nCandidates) {
        const heightToFitAllCandidates = this.candidateViewHeight * nCandidates;
        const height = Math.min(heightToFitAllCandidates, this._availableHeight, this.candidateViewHeight * 7 /* max # of candidates we want to show at once */);
        return height;
    }
    setCandidates(candidates) {
        const height = this._pickListHeight(candidates.length);
        this._listWidget.splice(0, 0, candidates);
        this._listWidget.layout(height);
        this._listContainer.style.height = `${height}px`;
    }
    clearCandidates() {
        this._listContainer.style.height = '0px';
        this._listWidget.splice(0, this._listWidget.length, []);
    }
    get focusedCandidate() {
        return this._listWidget.isDOMFocused() ? this._listWidget.getFocusedElements()[0].newSymbolName : undefined;
    }
    updateFont(fontInfo) {
        this._listContainer.style.fontFamily = fontInfo.fontFamily;
        this._listContainer.style.fontWeight = fontInfo.fontWeight;
        this._listContainer.style.fontSize = `${fontInfo.fontSize}px`;
        this._lineHeight = fontInfo.lineHeight;
        this._listWidget.rerender();
    }
    focusNext() {
        if (this._listWidget.isDOMFocused()) {
            this._listWidget.focusNext();
        }
        else {
            this._listWidget.domFocus();
            this._listWidget.focusFirst();
        }
        this._listWidget.reveal(this._listWidget.getFocus()[0]);
    }
    /**
     * @returns true if focus is moved to previous element
     */
    focusPrevious() {
        this._listWidget.domFocus();
        const focusedIx = this._listWidget.getFocus()[0];
        if (focusedIx !== 0) {
            this._listWidget.focusPrevious();
            this._listWidget.reveal(this._listWidget.getFocus()[0]);
        }
        return focusedIx > 0;
    }
}
export class CandidateView {
    constructor(parent, { lineHeight }) {
        this.domNode = document.createElement('div');
        this.domNode.style.display = `flex`;
        this.domNode.style.alignItems = `center`;
        this.domNode.style.height = `${lineHeight}px`;
        this.domNode.style.padding = `${CandidateView._PADDING}px`;
        this._icon = document.createElement('div');
        this._icon.style.display = `flex`;
        this._icon.style.alignItems = `center`;
        this._icon.style.width = this._icon.style.height = `${lineHeight * 0.8}px`;
        this.domNode.appendChild(this._icon);
        this._label = document.createElement('div');
        this._icon.style.display = `flex`;
        this._icon.style.alignItems = `center`;
        this._label.style.marginLeft = '5px';
        this.domNode.appendChild(this._label);
        parent.appendChild(this.domNode);
    }
    set model(value) {
        var _a;
        // @ulugbekna: a hack to always include sparkle for now
        const alwaysIncludeSparkle = true;
        // update icon
        if (alwaysIncludeSparkle || ((_a = value.tags) === null || _a === void 0 ? void 0 : _a.includes(NewSymbolNameTag.AIGenerated))) {
            if (this._icon.children.length === 0) {
                this._icon.appendChild(renderIcon(Codicon.sparkle));
            }
        }
        else {
            if (this._icon.children.length === 1) {
                this._icon.removeChild(this._icon.children[0]);
            }
        }
        this._label.innerText = value.newSymbolName;
    }
    static getLayoutInfo({ lineHeight }) {
        const totalHeight = lineHeight + CandidateView._PADDING * 2 /* top & bottom padding */;
        return { totalHeight };
    }
    dispose() {
    }
}
// TODO@ulugbekna: accessibility
CandidateView._PADDING = 2;
